#!/bin/sh

set -e -o pipefail
set -x

SCRIPT="$0"
SCRIPT_PWD="$PWD"
ROOT=$(dirname "$(readlink -f "$0")")

run_background_task() {
	cd "$SCRIPT_PWD"
	echo "Creating background task for '$1'. Log can be found in 'dmesg'."
	bos-tools daemonize /bin/sh -- \
		-c "source $SCRIPT; $1" "$SCRIPT" "--background"
	return 0
}

stop_and_kill_service() {
	service_name=$1
	process_name=$2
	/etc/init.d/$service_name stop &
	sleep 1
	killall -9 $process_name 2> /dev/null || true
}

file_size() {
    printf "0x%x" $(stat -c "%s" "$1")
}

rmdir_empty() {
	rmdir --ignore-fail-on-non-empty "$1"
}

# Supported architecture
CPU_ARCH_NEEDED="armv7l"
HW_MODEL_NAME="Xilinx Zynq"

# Supported NAND layout:
# dev:    size   erasesize  name
# mtd0: 02800000 00020000 "BOOT.bin-dts-marker-kernel"
# mtd1: 02000000 00020000 "ramfs"
# mtd2: 00800000 00020000 "configs"
# mtd3: 00200000 00020000 "sig"
# mtd4: 00e00000 00020000 "reserve1"
# mtd5: 01000000 00020000 "upgrade-ramfs"
# mtd6: 03800000 00020000 "upgrade-file"
# mtd7: 05800000 00020000 "reserve2"
PROC_MTD_SHA256="7aca7b8ac3114fd261b219c9b454025bd2e28fc786280f4eb146f27550facb93"

# Free space required on /tmp in KiBs
# The size corresponds to size of extracted rootfs (25004)
# + some extra space for future extensions
REQUIRED_FREE_SPACE_TMP=30720
SIZE_OF_CONFIG=$(df | grep '/config' | awk '{print $4}')
# It must comfortably fit 2x content of stock `config` partition and `nand_env`
REQUIRED_FREE_SPACE_TMP=$((REQUIRED_FREE_SPACE_TMP + 2 * $SIZE_OF_CONFIG))

LIB_LD_PATH="/lib/ld-musl-armhf.so.1"

POST_INSTALL_DIR="post_install"
POST_INSTALL_RUNNER="run.sh"
POST_INSTALL_STOCK_CONFIG="stock_config.json"
POST_INSTALL_PACKAGES_DIR="packages"
POST_INSTALL_TAR_PATH="$ROOT/post_install.tar.gz"

# Firmware images
FACTORY_PATH="$ROOT/factory.tar.gz"
MINER_CFG_IMAGE_PATH="$ROOT/miner_cfg.bin"
MINER_CFC_CONFIG_PATH="$ROOT/miner_cfg.config"
POST_INSTALL_SCRIPT_PATH="$ROOT/post_install.sh"
SYSTEM_BIT_PATH="$ROOT/system.bit.gz"
UIMAGE_PATH="$ROOT/uImage.itb"
UNSQUASHFS_PATH="$ROOT/unsquashfs"

UIMAGE_BACKUP_PATH="$ROOT/uImage_backup.gz"

FACTORY_PACKAGE_DIR="factory"
FACTORY_ROOT_REL_PATH="$FACTORY_PACKAGE_DIR/root"
FIRMWARE_ROOTFS_PATH="$ROOT/rootfs"

# Mount points
RESERVE1_MNT="/tmp/miner"
RESERVE1_UBI=2
RESERVE2_MNT="/nvdata"
RESERVE2_UBI=3

# NAND partitions
NAND_PART_BOOT_MTD=/dev/mtd0
NAND_PART_RESERVE1_MTD=/dev/mtd4
NAND_PART_RESERVE2_MTD=/dev/mtd7

# NAND images
NAND_UIMAGE_MTD=$NAND_PART_BOOT_MTD
NAND_UIMAGE_OFF=0x2000000
NAND_UIMAGE_LEN=0x800000
NAND_UIMAGE_CNT=64

NAND_SYSTEM_BIT_MTD=$NAND_PART_RESERVE1_MTD
NAND_SYSTEM_BIT_OFF=0

NAND_UIMAGE_BACKUP_MTD=$NAND_PART_RESERVE1_MTD
NAND_UIMAGE_BACKUP_OFF=0x100000

NAND_POST_INSTALL_HEADER_MTD=$NAND_PART_RESERVE2_MTD
NAND_POST_INSTALL_HEADER_OFF=0x100000

NAND_POST_INSTALL_TAR_MTD=$NAND_PART_RESERVE2_MTD
NAND_POST_INSTALL_TAR_OFF=0x200000

NAND_FACTORY_HEADER_MTD=$NAND_PART_RESERVE2_MTD
NAND_FACTORY_HEADER_OFF=0x3a00000

NAND_FACTORY_IMAGE_MTD=$NAND_PART_RESERVE2_MTD
NAND_FACTORY_IMAGE_OFF=0x3b00000

NAND_UBOOT_ENVS_MTD=$NAND_PART_BOOT_MTD
NAND_UBOOT_ENVS_OFF=0x1800000
NAND_UBOOT_ENVS_CNT=8

NAND_MINER_CFG_MTD=$NAND_UBOOT_ENVS_MTD
NAND_MINER_CFG_OFF=0x18a0000

get_net_config() {
	sed -n '/'$1'=/s/.*=["]*\([^"]*\)["]*/\1/p' "/config/network.conf"
}

create_file_size_header() {
	printf "0: %.8x" $(file_size "$1") | xxd -r -g0
}

# Check we're running on the right platform and all prerequisites are met
preinstall_check() {
	local file_list

    # Check we are not running BOS already
    if [ -f "/etc/bos_version" ]; then
        echo "Error: BOS is already installed" >&2
        return 1
    fi

    # Check required files are available
    file_list="$file_list $FACTORY_PATH"
    file_list="$file_list $MINER_CFG_IMAGE_PATH"
    file_list="$file_list $MINER_CFC_CONFIG_PATH"
    file_list="$file_list $POST_INSTALL_SCRIPT_PATH"
    file_list="$file_list $SYSTEM_BIT_PATH"
    file_list="$file_list $UIMAGE_PATH"
    file_list="$file_list $UNSQUASHFS_PATH"

    for i in $file_list; do
        if [ ! -f "$i" ]; then
            echo "Error: file $i required for installation is missing." >&2
            return 1
        fi
    done

    local tmp_free_space=$(df | grep 'tmpfs.*/tmp' | awk '{print $4}')
    if [ $REQUIRED_FREE_SPACE_TMP -gt "$tmp_free_space" ];  then
        echo "Error: Insufficient space on /tmp." >&2
        exit 1
    fi

    # Check NAND layout
    if ! sha256sum /proc/mtd | grep -q "$PROC_MTD_SHA256"; then
        echo "Error: Unsupported NAND layout" >&2
        cat /proc/mtd >&2
        sha256sum /proc/mtd
        return 1
    fi

    # Check architecture
    if ! uname -a | grep -q "$CPU_ARCH_NEEDED"; then
        echo "Error: Unsupported architecture (must be $CPU_ARCH_NEEDED)." >&2
        return 1
    fi

    # Check that the control board is Antminer
    if ! grep -q Antminer "/usr/bin/compile_time"; then
        echo "Error: This install script is only for Antminer." >&2
        exit 1
    fi

    # check board model name
    if ! cat /proc/device-tree/model | grep -q "$HW_MODEL_NAME"; then
        echo "Unsupported HW MODEL (must be ${HW_MODEL_NAME})" >&2
        return 1
    fi

	return 0
}

install_prepare() {
	tar -xzvf "$FACTORY_PATH" "$FACTORY_ROOT_REL_PATH"

	$UNSQUASHFS_PATH -d "$FIRMWARE_ROOTFS_PATH" "$FACTORY_ROOT_REL_PATH" \
		"/lib/libc.so" \
		"/lib/ld-musl-armhf.so.1" \
		"/usr/sbin/fw_printenv" \
		"/usr/sbin/fw_setenv" \
		"/usr/sbin/flash_erase" \
		"/usr/sbin/nanddump" \
		"/usr/sbin/nandwrite" \
		"/usr/sbin/ubidetach" \
		"/usr/bin/bos-tools" \
		"/usr/bin/killall" \
		"/usr/bin/md5sum" \
		"/usr/bin/xxd" \
		"/usr/bin/tr" \
		"/bin/busybox" \
		"/bin/umount" \
		"/bin/base64" \
		"/bin/dd" \
		"/bin/tar" \
		"/bin/gzip" \
		"/bin/sleep" \
		"/bin/sync" \
		"/bin/stat"

	rm -rf "$ROOT/$FACTORY_PACKAGE_DIR"

	# Create symlink in temporary '/lib/' directory which will be automatically
	# removed during restart (the rootfs is not persistent)
	ln -s "$FIRMWARE_ROOTFS_PATH/lib/ld-musl-armhf.so.1" "$LIB_LD_PATH"
	return 0
}

env_setup() {
	export PATH="$FIRMWARE_ROOTFS_PATH/usr/sbin:$FIRMWARE_ROOTFS_PATH/usr/bin:$FIRMWARE_ROOTFS_PATH/bin:$PATH"
	return 0
}

prepare_post_install_tar() {
	# Create post_install directory
	mkdir -p "$ROOT/$POST_INSTALL_DIR"
	# Create stock configuration and put it in post-install directory
	bos-tools create-stock-conf -o "$ROOT/$POST_INSTALL_DIR/$POST_INSTALL_STOCK_CONFIG"

	# Check if post-install have ipk packages
	if [ -d "$ROOT/$POST_INSTALL_DIR/$POST_INSTALL_PACKAGES_DIR" ]; then
		# Remove all non `.ipk` files
		cd "$ROOT/$POST_INSTALL_DIR/$POST_INSTALL_PACKAGES_DIR"
		# Subcommand can't be in quote to delete each file
		rm -rf $(ls | grep -v '\.ipk')
		cd -
		rmdir_empty "$ROOT/$POST_INSTALL_DIR/$POST_INSTALL_PACKAGES_DIR"
	fi

	# Check if post-install directory is non-empty and store it in NAND
	rmdir "$ROOT/$POST_INSTALL_DIR" 2>/dev/null && return 0 || true
	# Include default runner script
	mv "$POST_INSTALL_SCRIPT_PATH" "$ROOT/$POST_INSTALL_DIR/$POST_INSTALL_RUNNER"

	# Create post-install tarball
	tar -czvf "$POST_INSTALL_TAR_PATH" -C "$ROOT" "$POST_INSTALL_DIR"
	return 0
}

write_miner_cfg() {
	local ethaddr miner_hwid
	local hostname ipaddress netmask gateway dnsservers

	ethaddr=$(cat /sys/class/net/eth0/address)
	# Improve random seed with MAC address which should be unique
	echo ${ethaddr} > /dev/urandom
	miner_hwid=$(dd if=/dev/urandom bs=1 count=12 2>/dev/null | base64 | tr "+/" "ab")

	# Get stock network settings
	hostname=$(get_net_config "hostname")
	ipaddress=$(get_net_config "ipaddress")
	netmask=$(get_net_config "netmask")
	gateway=$(get_net_config "gateway")
	dnsservers=$(get_net_config "dnsservers" | tr " " ,)

	# Generate factory configuration for U-Boot
	nandwrite -p -s $NAND_MINER_CFG_OFF $NAND_MINER_CFG_MTD "$MINER_CFG_IMAGE_PATH"
	fw_setenv -c "$MINER_CFC_CONFIG_PATH" --script - <<-EOF
		ethaddr ${ethaddr}
		miner_hwid ${miner_hwid}
		# Network settings
		net_hostname ${hostname}
		net_ip ${ipaddress}
		net_mask ${netmask}
		net_gateway ${gateway}
		net_dns_servers ${dnsservers}
	EOF
	fw_printenv -c "$MINER_CFC_CONFIG_PATH"
}

write_essentials() {
    # Stop daemons
	stop_and_kill_service S50lighttpd lighttpd
	stop_and_kill_service S71monitorcg S71monitorcg
	stop_and_kill_service S70cgminer cgminer
	stop_and_kill_service S70cgminer bmminer
	stop_and_kill_service S71monitor-cminer S71monitor-cminer
	stop_and_kill_service S70cminer cminer 50
	stop_and_kill_service S52miner_act S52miner_act
	stop_and_kill_service S66monitor-recobtn monitor-recobtn
	stop_and_kill_service S01syslogd syslogd
	stop_and_kill_service S41ntp ntpd
	stop_and_kill_service S50avahi-daemon avahi-daemon

	# Detach all "reserve" partitions
	umount "$RESERVE1_MNT"
	ubidetach /dev/ubi_ctrl -d $RESERVE1_UBI
	umount "$RESERVE2_MNT"
	ubidetach /dev/ubi_ctrl -d $RESERVE2_UBI
	sync

	# Backup uImage
	nanddump -s $NAND_UIMAGE_OFF -l $NAND_UIMAGE_LEN $NAND_UIMAGE_MTD \
	| gzip > "$UIMAGE_BACKUP_PATH"

	# Erase the whole "reserve1" partition
	flash_erase $NAND_PART_RESERVE1_MTD 0 0
	# Write compressed FPGA bitstream
	nandwrite -p -s $NAND_SYSTEM_BIT_OFF $NAND_SYSTEM_BIT_MTD "$SYSTEM_BIT_PATH"
	# Write compressed backup of uImage
	nandwrite -p -s $NAND_UIMAGE_BACKUP_OFF $NAND_UIMAGE_BACKUP_MTD "$UIMAGE_BACKUP_PATH"

	# Erase the whole "reserve2" partition
	flash_erase $NAND_PART_RESERVE2_MTD 0 0
	# Write size of factory image to the first block of factory partition
	create_file_size_header "$FACTORY_PATH" | \
	nandwrite -p -s $NAND_FACTORY_HEADER_OFF $NAND_FACTORY_HEADER_MTD -
	# Write factory image to factory partition after header block
	nandwrite -p -s $NAND_FACTORY_IMAGE_OFF $NAND_FACTORY_IMAGE_MTD "$FACTORY_PATH"
	# Create size for post-install tarball
	rm "$FACTORY_PATH"

	# Erase uboot_env+miner_cfg
	flash_erase $NAND_UBOOT_ENVS_MTD $NAND_UBOOT_ENVS_OFF $NAND_UBOOT_ENVS_CNT
	write_miner_cfg || exit 1

	# Detect post-install tarball or finish the writing
	[ -f "$POST_INSTALL_TAR_PATH" ] || return 0

	# Write size of post-install tarball to the first block of firmware2 partition
	create_file_size_header "$POST_INSTALL_TAR_PATH" | \
	nandwrite -p -s $NAND_POST_INSTALL_HEADER_OFF $NAND_POST_INSTALL_HEADER_MTD -
	# Write post-install upgrade tarball to firmware2 partition after header block
	nandwrite -p -s $NAND_POST_INSTALL_TAR_OFF $NAND_POST_INSTALL_TAR_MTD \
		"$POST_INSTALL_TAR_PATH"
	return 0
}

install_uimage() {
	# This part of installation is critical and is run on background
	# to be completed even when SSH session is terminated
	env_setup

	# Wait for first part of script to finish,
	# because killall of dropbear also kills child part of this script
	# if parent is still running
	sleep 1

	# Terminate dropbear
	stop_and_kill_service S50dropbear dropbear

	# Write new uImage with U-Boot and recovery FIT image
	flash_erase $NAND_UIMAGE_MTD $NAND_UIMAGE_OFF $NAND_UIMAGE_CNT
	nandwrite -p -s $NAND_UIMAGE_OFF /dev/mtd0 "$UIMAGE_PATH"

	sync
	reboot
	exit 0
}

trap_handler() {
    local rc=$?

    # Remove the temp directory only if ends with error
    # In case of end with success, the temp directory needs to be kept
    # for second stage of installation
    [ "$rc" -eq 0 ] && exit 0

    echo "Removing temp install directory"
    rm -rf "$ROOT"
	# Remove also symlink to LD library from stock system
	rm -f "$LIB_LD_PATH"

    exit $rc
}

# Stop execution when the script is running in the background
if [ "$1" == "--background" ]; then
	return 0
fi

# Add `trap_handler` to remove temp install directory
trap "trap_handler" EXIT

preinstall_check || exit 1
install_prepare || exit 1
env_setup || exit 1
prepare_post_install_tar || exit 1
write_essentials || exit 1

run_background_task install_uimage
